#include <math.h>
#include <assert.h>
#include "arc.h"


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////  ArcVertex  ///////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

ArcVertex::ArcVertex( Vertex &x, VertexAVL &tree )
    : Vertex( x ), m_tree( tree )
{
    // cout << "ArcVertex::ArcVertex: " << (void *)this << endl;
    m_count = 0;
}


ArcVertex::~ArcVertex()
{
    // cout << "ArcVertex::~ArcVertex: " << (void*)this << endl;
}


// FIXME:  The whole AVL thing is full of crap.  It doesn't guarantee
// anything.  Damn!
void
ArcVertex::deleteMaybe()
{
    assert( m_count == 0 );
    ArcVertex *result = (ArcVertex *) m_tree.remove( this );
    if( result == this )
	delete this;
}

/////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Arc  //////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

Arc::Arc( ArcVertex *v1, ArcVertex *v2 )
{
    construct( v1, v2 );
}


Arc::Arc( ArcVertex &v1, ArcVertex &v2 )
{
    construct( &v1, &v2 );
}


Arc::Arc( const Arc &a )
{
    construct( a.m_v1, a.m_v2 );
}


Arc::~Arc()
{
    /*
    cout << "Arc::~Arc: " << (void*)this << ": (" << (void*)m_v1
	 << ',' << (void*) m_v2 << ")  (" << m_v1->count() << ','
	 << m_v2->count() << ")\n";
	 */
    if( m_normal )
	delete m_normal;
    assert( m_v1->count() > 0 );
    assert( m_v2->count() > 0 );
    if( !m_v1->decrement() )
	m_v1->deleteMaybe();
    if( !m_v2->decrement() )
	m_v2->deleteMaybe();
}



void
Arc::construct( ArcVertex *v1, ArcVertex *v2 )
{
    /*
    cout << "Arc::Arc: " << (void*)this << ": (" << (void*)v1 << ','
	 << (void*)v2 << ")  (" << v1->count() << ',' << v2->count() << ")\n";
    */
    m_v1 = v1;
    m_v2 = v2;
    m_v1->increment();
    m_v2->increment();
    Vertex d = Vertex::linkedTriad( *m_v1, -1, *m_v2 );
    m_length2 = d.mag2();
    m_normal = 0;
}


Vertex &
Arc::getGreatCircle() 
{
    if( !m_normal )
    {
	m_normal = new Vertex( Vertex::crossProduct(m_v1,m_v2) );
	m_normal->normalize();
    }
    return *m_normal;
}


// "inside" means on the right-handed side of the arc (where the fingers
//  indicate the direction of the arc)
int
Arc::inOrOut( Vertex &v )
{
    getGreatCircle();

    double d = Vertex::dotProduct( *m_normal, v );
    if( d > EPSILON )
	return 1;  // it's inside
    if( d < -EPSILON )
	return -1;
    return 0;
}


Vertex
Arc::intersection( Vertex &v, Arc &a )
{
    Vertex &g = a.getGreatCircle();
    Vertex tangent = Vertex::crossProduct( v, g );

    tangent.normalize();
    if( !a.between(tangent) )
    {
	tangent.scale( -1 );
	assert( a.between(tangent) );
    }
    assert( tangent.v[0]!=0 || tangent.v[1]!=0 || tangent.v[2]!=0 );
    return tangent;
}


Vertex
Arc::intersection( Arc &a1, Arc &a2 )
{
    Vertex &g1 = a1.getGreatCircle();
    Vertex &g2 = a2.getGreatCircle();
    Vertex tangent = Vertex::crossProduct( g1, g2 );
    Vertex answer;

    if( tangent.mag2() > EPSILON )
    {
	if( a1.between(tangent) )
	{
	    if( a2.between(tangent) )
		answer = tangent;
	}
	else
	{
	    tangent.scale( -1 );
	    if( a1.between(tangent) && a2.between(tangent) )
		answer = tangent;
	}
    }
    return answer;
}


int
Arc::coincident( Arc &a, Arc &b )
{
    double d = Vertex::dotProduct( a.getGreatCircle(), b.getGreatCircle() );
    d = fabs( d );
    return d >= (1-EPSILON);
}


// Let t.c lie on the chord (v1,v2).  If t>=0 then c is between v1 and v2.
// note:  implicit equation of chord is n.x = n.v1
int
Arc::between( Vertex &c )
{
    Vertex n = Vertex::linkedTriad( m_v1, 1, m_v2 );
    double d = Vertex::dotProduct( n, *m_v1 );
    double denom = Vertex::dotProduct( c, n );
    if( fabs(denom) < EPSILON )
	return 1;
    else
	return d/denom >= 0;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////  VertexAVL / //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

ArcVertex *
VertexAVL::getVertex( Vertex &v )
{
    assert( fabs(v.mag2()-1) < 0.05 );
    ArcVertex *s = (ArcVertex *) search( &v );
    if( !s )
    {
	s = new ArcVertex( v, *this );
	insert( s );
    }
    return s;
}


/////////////////////////////////////////////////////////////////////////////
///////////////////////////////  Triangle  //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


Triangle::Triangle( Arc *a, Arc *b, Arc *c )
    : m_arcs( 3 )
{
    m_arcs.add( a );
    m_arcs.add( b );
    m_arcs.add( c );
}


Triangle::~Triangle()
{
    m_arcs.clear( 1 );
}


void Triangle::dismantle()
{
    m_arcs.clear();
}


Vertex Triangle::normal()
{
    Vertex normal;
    for( int i=0; i<3; i++ )
    {
	int j = (i+1) % 3;
	double *vi = m_arcs[i]->getV1().v;
	double *vj = m_arcs[j]->getV1().v;
	normal.v[0] += ( vi[1] - vj[1] )*( vi[2] + vj[2] );
	normal.v[1] += ( vi[2] - vj[2] )*( vi[0] + vj[0] );
	normal.v[2] += ( vi[0] - vj[0] )*( vi[1] + vj[1] );
    }
    normal.normalize();
    return normal;
}


// argument is consumed
DArray<Triangle> Triangle::triangulate( DArray<ArcVertex> &array )
{
    assert( array.length() >= 3 );
    DArray<Triangle> newTriangles;
    int i = 1;
    int j = array.length() - 1;
    int k = 0;
    int parity = 0;
    do
    {
	// add new triangle to list
	Triangle *t = new Triangle(
	    new Arc(array[k],array[i]),
	    new Arc(array[i],array[j]),
	    new Arc(array[j],array[k]) );
	newTriangles.add( t );

	// update pointers
	k = parity ? i++ : j--;
	parity = !parity;
    } while( i != j );
    array.clear();
    return newTriangles;
}


// argument is consumed
DArray<Triangle> Triangle::triangulate( DArray<Arc> &a )
{
    assert( a.length() >= 3 );

    DArray<Triangle> out;
    if( a.length() == 3 )
	out.add( new Triangle(a[0],a[1],a[2]) );
    else
    {
	int i;
	ArcVertex &pivot = a[0]->getV1();

	out.add( new Triangle(a[0],a[1],new Arc(a[1]->getV2(),pivot)) );
	for( i=2; i<a.length()-2; i++ )
	{
	    Triangle *t = new Triangle( new Arc(pivot,a[i]->getV1()),
		a[i], new Arc(a[i]->getV2(),pivot) );
	    out.add( t );
	}
	out.add( new Triangle(a[i],a[i+1],new Arc(pivot,a[i]->getV1())) );
    }
    a.clear();
    return out;
}


ostream &operator<<( ostream &out, Triangle &t )
{
    DArray<Arc> arcs = t.arcs();

    out << "Triangle: ";
    for( int i=0; i<arcs.length(); i++ )
    {
	out << arcs[i]->getV1();
	if( i+1 < arcs.length() )
	    out << ", ";
    }
    return cout << '\n';
}

