#include <math.h>
#include <values.h>
#include "base.h"

/////////////////////////////////////////////////////////////////////////////
//////////////////////////////  Vertex  /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

Vertex Vertex::crossProduct( const Vertex &a, const Vertex &b )
{
    Vertex c(
	a[1]*b[2] - b[1]*a[2],
	b[0]*a[2] - a[0]*b[2],
	a[0]*b[1] - b[0]*a[1] );
    return c;
}


double Vertex::dotProduct( const Vertex &a, const Vertex &b )
{
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}


double Vertex::normalize()
{
    double m = mag();
    scale( 1/m );
    return m;
}



Vertex Vertex::linkedTriad( const Vertex &a, double s, const Vertex &b )
{
    Vertex c( a[0] + s*b[0], a[1] + s*b[1], a[2] + s*b[2] );
    return c;
}


ostream &operator<<( ostream &out, const Vertex &v )
{
    out << '(' << v[0] << ',' << v[1] << ',' << v[2] << ')';
    return out;
}


/////////////////////////////////////////////////////////////////////////////
////////////////////////////  BoundingBox  //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


BoundingBox::BoundingBox( DArray<Polygon> &a )
{
    init();
    for( int i=0; i<a.length(); i++ )
	merge( a[i]->getVerts() );
    getCorners();
}


BoundingBox::BoundingBox( Polygon &p )
{
    init();
    merge( p.getVerts() );
    getCorners();
}


BoundingBox::BoundingBox( BoundingBox &a, BoundingBox &b )
{
    init();
    merge( a.min );
    merge( a.max );
    merge( b.min );
    merge( b.max );
    getCorners();
}

void
BoundingBox::init()
{
    min[0] = MAXDOUBLE;
    min[1] = MAXDOUBLE;
    min[2] = MAXDOUBLE;
    max[0] = MINDOUBLE;
    min[1] = MINDOUBLE;
    min[2] = MINDOUBLE;
}

void
BoundingBox::getCorners()
{
    static int m[][3] =
    {
	{ 1, 1, 1 },
	{ 1, 1, -1 },
	{ 1, -1, 1 },
	{ 1, -1, -1 },
	{ -1, 1, 1 },
	{ -1, 1, -1 },
	{ -1, -1, 1 },
	{ -1, -1, -1 }
    };
    
    Vertex c = min + max;
    c.scale( 0.5 );
    Vertex d = max - min;
    d.scale( 0.5 );
    
    for( int i=0; i<8; i++ )
    {
	corners.add( new Vertex(c[0]+d[0]*m[i][0],
				c[1]+d[1]*m[i][1],
				c[2]+d[2]*m[i][2]) );
    }
}

void
BoundingBox::merge( Vertex &v )
{
    for( int k=0; k<3; k++ )
    {
	if( v[k] < min[k] )
	    min[k] = v[k];
	else if( v[k] > max[k] )
	    max[k] = v[k];
    }
}

void
BoundingBox::merge( DArray<Vertex> &verts )
{
    for( int j=0; j<verts.length(); j++ )
	merge( *verts[j] );
}        


/////////////////////////////////////////////////////////////////////////////
//////////////////////////////  Polygon  ////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


Polygon::Polygon( DArray<Vertex> &array, int label )
{
    m_normal = 0;
    m_label = label;
    for( int i=0; i<array.length(); i++ )
	m_verts.add( new Vertex(*array[i]) );
}


Polygon::Polygon( Vertex &a, Vertex &b, Vertex &c, int label )
{
    m_normal = 0;
    m_label = label;
    m_verts.add( new Vertex(a) );
    m_verts.add( new Vertex(b) );
    m_verts.add( new Vertex(c) );
}


Polygon::~Polygon()
{
    if( m_normal )
	delete m_normal;
    m_verts.clear( 1 );
}


Vertex Polygon::center()
{
    int length = m_verts.length();
    Vertex c;

    c.v[0] = 0; c.v[1] = 0; c.v[2] = 0;
    for( int i=0; i<length; i++ )
    {
	c.v[0] += m_verts[i]->v[0];
	c.v[1] += m_verts[i]->v[1];
	c.v[2] += m_verts[i]->v[2];
    }
    c.scale( 1./length );
    return c;
}


Vertex &Polygon::normal()
{
    if( !m_normal )
    {
	m_normal = new Vertex();
	for( int i=0; i<m_verts.length(); i++ )
	{
	    int j = (i+1) % m_verts.length();
	    double *vi = m_verts[i]->v;
	    double *vj = m_verts[j]->v;
	    m_normal->v[0] += ( vi[1] - vj[1] )*( vi[2] + vj[2] );
	    m_normal->v[1] += ( vi[2] - vj[2] )*( vi[0] + vj[0] );
	    m_normal->v[2] += ( vi[0] - vj[0] )*( vi[1] + vj[1] );
	}
	m_normal->normalize();

    }
    return *m_normal;
}


double Polygon::planeConstant()
{
    return Vertex::dotProduct( *m_verts[0], normal() );
}


ostream &operator<<( ostream &out, Polygon &p )
{
    out << "Polygon(" << p.m_label << "): ";
    for( int i=0; i<p.m_verts.length(); i++ )
    {
	out << *p.m_verts[i];
	if( i+1 <p.m_verts.length() )
	    out << ", ";
    }
    out << '\n';
    return out;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////  Miscellaneous  ///////////////////////////////
/////////////////////////////////////////////////////////////////////////////


int
compare( const Vertex &x, const Vertex &y )
{
    double diff;

    diff = x[0] - y[0];
    if( fabs(diff) > EPSILON )
	return diff<0 ? -1 : 1;
    diff = x[1] - y[1];
    if( fabs(diff) > EPSILON )
	return diff<0 ? -1 : 1;
    diff = x[2] - y[2];
    if( fabs(diff) > EPSILON )
	return diff<0 ? -1 : 1;
    return 0;
}

