#ifndef _BASE_H // -*-C++-*-
#define _BASE_H

#include <math.h>
#include <ostream.h>
#include <libtk/darray.h>

const double EPSILON = 1e-6;  // man, what a hack!

class Vertex
{
public:
    double v[3];

    Vertex()  { v[0]=0; v[1]=0; v[2]=0; }
    Vertex( const Vertex &x ) { v[0]=x[0]; v[1]=x[1]; v[2]=x[2]; }
    Vertex( double _v[3] ) { v[0]=_v[0]; v[1]=_v[1]; v[2]=_v[2]; }
    Vertex( float _v[3] ) { v[0]=_v[0]; v[1]=_v[1]; v[2]=_v[2]; }
    Vertex( double x, double y, double z ) { v[0]=x; v[1]=y; v[2]=z; }
    virtual ~Vertex() {};

    void   set( double a, double b, double c ) { v[0]=a; v[1]=b; v[2]=c; }
    double mag2() const { return v[0]*v[0]+v[1]*v[1]+v[2]*v[2]; }
    double mag() const { return sqrt( mag2() ); }
    double normalize();
    void scale( double s ) { v[0]*=s; v[1]*=s; v[2]*=s; }
    double operator[]( int i ) const { return v[i]; }
    double &operator[]( int i ) { return v[i]; }

    static Vertex crossProduct( const Vertex &, const Vertex & );
    static Vertex crossProduct( const Vertex *a, const Vertex *b ) 
	{ return crossProduct( *a, *b ); }
    static double dotProduct( const Vertex &, const Vertex & );
    static double dotProduct( const Vertex *a, const Vertex *b )
	{ return dotProduct( *a, *b ); }
    static Vertex linkedTriad( const Vertex &, double, const Vertex & );
    static Vertex linkedTriad( const Vertex *a, double x, const Vertex *b )
	{ return linkedTriad( *a, x, *b ); }

    friend ostream &operator<<( ostream &, const Vertex & );
    friend int      compare( const Vertex &, const Vertex & );

    // arithmetic
    friend Vertex operator+( const Vertex &a, const Vertex &b )
    {
	return Vertex::linkedTriad( a, 1, b );
    }
    friend Vertex operator-( const Vertex &a, const Vertex &b )
    {
	return Vertex::linkedTriad( a, -1, b );
    }
    friend Vertex operator-( const Vertex &a )
    {
	Vertex b = a; b.scale( -1 ); return b;
    }
    friend double operator*( const Vertex &a, const Vertex &b )
    {
	return Vertex::dotProduct( a, b );
    }
    friend Vertex operator^( const Vertex &a, const Vertex &b )
    {
	return Vertex::crossProduct( a, b );
    }

    // more operators
    friend void operator+=( Vertex &a, const Vertex &b )
    {
	a = a + b;
    }
    friend void operator-=( Vertex &a, const Vertex &b )
    {
	a = a - b;
    }
};


class Polygon
{
    DArray<Vertex> m_verts;
    Vertex *m_normal;
    int m_label;

public:
    Polygon( DArray<Vertex> &vlist, int label=0 );
    Polygon( Vertex &, Vertex &, Vertex &, int label=0 );
    ~Polygon();

    int numVerts() { return m_verts.length(); }
    DArray<Vertex> &getVerts() { return m_verts; } // for reference only
    Vertex center();
    Vertex &normal();
    double planeConstant();
    int label() { return m_label; }

    friend ostream &operator<<( ostream &, Polygon & );
};


class BoundingBox
{
    void init();
    void getCorners();
    void merge( DArray<Vertex> & );
    void merge( Vertex & );
public:
    Vertex min;
    Vertex max;
    DArray<Vertex> corners;

    BoundingBox( DArray<Polygon> & );
    BoundingBox( BoundingBox &, BoundingBox & );
    BoundingBox( Polygon & );
    ~BoundingBox() { corners.clear( 1 ); }
};


#endif // _BASE_H
