#include <math.h>
#include <iostream.h>
#include <libtk/map.h>
#include "factor.h"
#include "io.h"



FormFactor::FormFactor( Camera &cam, DArray<Polygon> &src,
    DArray<Polygon> &occ, int storeSrc )
    : m_point(cam.m_eye)
{
    DArray<Polygon> total = src;
    Vertex lookAt( cam.m_lookAt );
    m_normal = lookAt - m_point; m_normal.normalize();
    
    total.append( occ );
    m_bsp = new BSP( total );
    m_sphere = new Sphere( cam );

    // add all sources into a map
    for( int i=0; i<src.length(); i++ )
    {
	Polygon *p = src[i];
	m_srcMap[p->label()] = new DArray<Polygon>;
    }

    // get polygons in front to back order
    DArray<Polygon> result = m_bsp->traverse( m_point );
    if( result.length() != total.length() )
	cout << "FormFactor:  " << total.length() << " polygons into BSP, "
	     << result.length() << " out.\n";

    // add them into the sphere, but store the results from sources
    for( int i=0; i<result.length(); i++ )
    {
	Polygon *p = result[i];
	int isSource = m_srcMap.contains( p->label() );

	// cout << "\nFormFactor: " << (isSource?"source: ":"occluder: ") << *p;

	if( i % 10 == 0 )
	{
	    cerr << ' ' << i;
	    cerr.flush();
	}
	
	if( !isSource )  // it's just an occluder
	{
	    m_sphere->renderOccluder( *p );
	}
	else
	{
	    DArray<Polygon> *source = m_srcMap[p->label()];
	    DArray<Polygon> visible = m_sphere->getVisibleSource( *p,
								  storeSrc );
	    source->append( visible );
	}
    }
    result.clear( 1 );  // don't need these any more
}


DArray<Polygon> &
FormFactor::getRemainder( Polygon &p )
{
    static DArray<Polygon> empty;
    DArray<Polygon> *ret;

    if( m_srcMap.contains(p.label()) )
	ret = m_srcMap[p.label()];
    else
	ret = &empty;
    return *ret;
}


// there's some room for optimization here
double
FormFactor::getFormFactor( Polygon &p )
{

    double total = 0;

    if( !m_srcMap.contains(p.label()) )
    {
	cerr << "FormFactor::getFormFactor:  bad label: " << p.label() << endl;
	return -1;
    }

    DArray<Polygon> &src = *m_srcMap[p.label()];
    for( int i=0; i<src.length(); i++ )
    {
	DArray<Vertex> &list = src[i]->getVerts();

	for( int j=0; j<list.length(); j++ )
	{
	    int k = (j+1) % list.length();
	    Vertex R1 = Vertex::linkedTriad( *list[j], -1, m_point );
	    Vertex R2 = Vertex::linkedTriad( *list[k], -1, m_point );
	    R1.normalize();
	    R2.normalize();
	    Vertex g = Vertex::crossProduct( R2, R1 );
	    double theta = asin( g.mag() );
	    total += Vertex::dotProduct( m_normal, g )*theta/g.mag();
	}
    }
    total *= 1/(2*M_PI);
    return total;
}


FormFactor::~FormFactor()
{
    delete m_bsp;
    delete m_sphere;

    DArray<int> keys = m_srcMap.keys();
    for( int i=0; i<keys.length(); i++ )
    {
	DArray<Polygon> *array = m_srcMap[*keys[i]];
	array->clear( 1 );
	delete array;
    }
}


ostream &
operator<<( ostream &out, FormFactor &ff )
{
    DArray<Triangle> triangles = ff.m_sphere->getTriangles();
    DArray<Polygon> polygons;

    if( triangles.length() == 0 )
	cerr << "Warning: FormFactor operator<<:  no triangles!\n";
    
    for( int i=0; i<triangles.length(); i++ )
    {
	Triangle &t = *triangles[i];
	Vertex v1 = t.getVertex( 0 );
	Vertex v2 = t.getVertex( 1 );
	Vertex v3 = t.getVertex( 2 );
	polygons.add( new Polygon(v1,v2,v3) );
    }
    out << "# Output from FormFactor:\n";
    out << polygons;
    out << "# End of FormFactor\n";
    polygons.clear( 1 ); // unallocate
    return out;
}
    
