#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>
#include <libtk/darray.h>
#include <libtk/camera.h>
#include "factor.h"
#include "hemi.h"
#include "jfobj.h"
#include "nfobj.h"

#include "io.h"

static void usage();
static void initializeCamera( Camera &, const char * );
static DArray<Polygon> readFile( char * );

int
main( int argc, char **argv )
{
    Camera camera;
    int doHemi = 0;
    int doFF = 1;
    char *inputFile = 0;
    char *camFile = 0;
    int printFF = 1;
    int writeOutput = 1;
    int writeInput = 0;

    if( argc < 2 )
	usage();
    for( int i=1; i<argc; i++ )
    {
	if( argv[i][0] == '-' )
	{
	    switch( argv[i][1] )
	    {
	    case 'c':
		camFile = argv[++i];
		break;
	    case 'f':
		doFF = 0;
		break;
	    case 'h':
		doHemi = 1;
		break;
	    case 'i':
		writeInput = 1;
		break;
	    case 'n':
		writeOutput = 0;
		break;
	    case 'p':
		printFF = 0;
		break;
	    default:
		cerr << "Unrecognized option: " << argv[i] << endl;
		exit( 1 );
	    }
	}
	else
	{
	    if( !inputFile )
		inputFile = argv[i];
	    else
		usage();
	}
    }
    if( !inputFile )
	usage();

    initializeCamera( camera, camFile );
    DArray<Polygon> occluders;
    DArray<Polygon> sources = readFile( inputFile );
    BoundingBox bb( sources );
    cout << "Read " << sources.length() << " polygons.\n";

    if( writeInput )
	write( sources, "input.obj" );

    if( !doFF )
	exit( 0 );
    
    // assert( sources.length() > 1 );

    FormFactor ff( camera, sources, occluders, 1 ); // store sources

    DArray<Polygon> results;
    double total = 0;
    int numTriangles = 0;
    for( int i=0; i<sources.length(); i++ )
    {
	Polygon &p = *sources[i];
	DArray<Polygon> &out = ff.getRemainder( p );
	// assert( out.length() > 0 );  this can legally happen
	double factor = ff.getFormFactor( p );
	if( printFF )
	{
	    cout << "Source " << p.label() << ":  " << out.length() 
		 << " triangles, ff: " << ff.getFormFactor( p ) << endl;
	}
	results.append( out );
	total += factor;
	numTriangles += out.length();
    }
    cout << "Total form factor: " << total << endl;
    cout << "Total number of triangles: " << numTriangles << endl;


    /*
    if( !write(ff,"sphere.obj") )
	cerr << "Cannot write sphere.obj\n";
	*/

    if( writeOutput )
	write( results,  "out.obj" );

    if( doHemi )
    {
	DArray<Vertex> &corners = bb.corners;
	Vertex eye( camera.m_eye );
	Vertex lookAt( camera.m_lookAt );
	Vertex view = lookAt - eye;
	view.normalize();

	double near = MAXDOUBLE;
	double far = MINDOUBLE;
	for( int i=0; i<corners.length(); i++ )
	{
	    Vertex &c = *corners[i];
	    Vertex x = c - eye;
	    double d = view * x;
	    if( d < near )
		near = d;
	    else if( d > far )
		far = d;
	}
	near *= 0.5;
	far *= 2;
	cout << "near:  " << near << "  far: " << far << endl;
	if( near <= 0 || far < near )
	{
	    cerr << "Warning: Bad camera for this scene: "
		"near: " << near << "  far: " << far << endl;
	    near = -1;
	    far = -1;
	}
	HemiCube h( camera, sources, occluders, near, far, 0 );
	h.formFactor(); // this never returns, unless windows are hidden
    }
    cout << "Done.\n";
    exit( 0 );
    return 0;
}


void
usage()
{
    cout << "Usage:  ff [options] inFile\n";
    cout << "   -c camFile      Specify a camera file\n";
    cout << "   -f              Don't do form factors\n";
    cout << "   -h              Do hemicube\n";
    cout << "   -i              Write input file (as OBJ)\n";
    cout << "   -n              Don't write output file\n";
    cout << "   -p              Don't print form factor info\n";
    exit( 1 );
}
	

void
initializeCamera( Camera &cam, const char *file )
{
    if( file )
	cam.read( file );
    else
    {
	cam.m_eye[0] = 0; cam.m_eye[1] = 0; cam.m_eye[2] = 0;
	cam.m_lookAt[0] = 0; cam.m_lookAt[1] = 0; cam.m_lookAt[2] = 1;
	cam.m_up[0] = 0; cam.m_up[1] = 1; cam.m_up[2] = 0;
	cam.m_fovY = 45;
	cam.m_width = 128;
	cam.m_height = 128;
    }
}
	

DArray<Polygon>
readFile( char *name )
{
    int len = strlen( name );

    if( len > 4 && !strcmp(name+(len-4),".obj") )
    {
	DArray<Polygon> array;
	if( !read(array,name) )
	    cerr << "Unable to read file: " << name << '\n';
	return array;
    }
    else if( len > 3 && !strcmp(name+(len-3),".jf") )
    {
	DArray<Polygon> array = JFFile::read( name );
	// write( array, "converted.obj" );
	return array;
    }
    else if( len >3 && !strcmp(name+(len-3),".nf") )
    {
 	DArray<Polygon> array = NFFile::read( name );
	return array;
    }
    else
    {
	DArray<Polygon> array;
	cout << "Unknown file extension, assuming OBJ.\n";
	if( !read(array,name) )
	    cerr << "Unable to read file: " << name << '\n';
	return array;
    }
}
	
