#include <assert.h>
#include <stdlib.h>
#include <iostream.h>
#include "sphere.h"


class ClipMatrix
{
    int m[2][3][3];  // hee, hee (variable's name)

public:
    ClipMatrix( Triangle &, Triangle & );
    int overlap();
};


static DArray<Triangle> subtract( VertexAVL &, Triangle &, Triangle & );
static DArray<Arc>      normalize( DArray<Arc> & );
static DArray<Arc>      crop( VertexAVL &, Arc &, DArray<Arc> & );
static DArray<Triangle> cullDegenerate( DArray<Triangle> & );

static void assertPolygon( DArray<Arc> & );
static void assertIntersection( Vertex &, Vertex &, Arc & );
static void assertConsistent( DArray<Triangle> & );

static ostream &operator<<( ostream &, DArray<Arc> & );
static ostream &operator<<( ostream &, Arc & );
static void myAssert( int );


Sphere::Sphere( Camera &cam )
    : m_point(cam.m_eye)
{
    Vertex lookAt( cam.m_lookAt );
    m_normal = lookAt - m_point;
    m_normal.normalize();
#ifdef USE_TRITREE
    m_tritree = new TriTree( m_normal, 3 );
#endif // USE_TRITREE    
}


Sphere::~Sphere()
{
    // delete m_triTree/triangles
#ifdef USE_TRITREE
    delete m_tritree;
#else    
    m_triangles.clear( 1 );
#endif // USE_TRITREE
    
    // FIXME:
    // There may be left over things in the tree, but they should
    // all have 0 reference counts.  See also the fixme in arc.C.
    // Anyway, we'll just delete them manually.
    DSArray<Vertex> items = m_vertTree.items();
    if( items.length() != 0 )
    {
	cout << items.length() << " ArcVertices undeleted\n";
	for( int i=0; i<items.length(); i++ )
	{
	    ArcVertex *v = (ArcVertex *) items[i];
	    assert( v->count() == 0 );
	    delete v;
	}
    }
}


void
Sphere::renderOccluder( Polygon &p )
{
    DArray<Triangle> array = render( p );
    insert( array );
}


DArray<Polygon>
Sphere::getVisibleSource( Polygon &p, int store )
{
    DArray<Triangle> triangles = render( p );
    DArray<Polygon> polys;

    /*
    cout << "Sphere::getVisibleSource:  render returned: \n";
    for( int i=0; i<triangles.length(); i++ )
	cout << "  " <<  *triangles[i];
    */

    assertConsistent( triangles );
    
    if( store )
	insert( triangles );



    if( triangles.length() != 0 )
    {
	Vertex &n = p.normal();
	double d = p.planeConstant();
	double numerator = d - Vertex::dotProduct( m_point, n );

	for( int i=0; i<triangles.length(); i++ )
	{
	    Triangle &tri = *triangles[i];
	    Vertex verts[3];
	    for( int j=2; j>=0; j-- ) // remember to reverse vertices
	    {
		Vertex &v = tri.getVertex( j );
		double t = numerator/Vertex::dotProduct(v,n);
		verts[2-j] = Vertex::linkedTriad( m_point, t, v );
	    }
	    polys.add( new Polygon(verts[0],verts[1],verts[2]) );
	}
    }

    /*
    cout << "Sphere::getVisibleSource: returning:\n";
    for( int i=0; i<polys.length(); i++ )
    {
	cout << "  " << *polys[i];
    }
    */

    for( int i=0; i<polys.length(); i++ )
    {
	int j = (i+1) % polys.length();
	Vertex &n1 = polys[i]->normal();
	Vertex &n2 = polys[j]->normal();
	myAssert( Vertex::dotProduct(n1,n2) > 0.5 );
    }

    if( !store )
	triangles.clear( 1 ); // unallocate

    return polys;
}


DArray<Triangle>
Sphere::getTriangles()
{
#ifdef USE_TRITREE
    return m_tritree->getAllTriangles();
#else
    return m_triangles;
#endif
}


DArray<Triangle>
Sphere::render( Polygon &p )
{
    DArray<Triangle> newTriangles;

    if( p.normal()*m_normal >= 0 )
    {
	cout << "Warning:  polygon not facing viewer\n";
	return newTriangles;
    }
    
    //// triangularize polygon
    // transform vertices
    DArray<Vertex> &verts = p.getVerts();
    DArray<ArcVertex> array;
    for( int i=verts.length()-1; i>=0; i-- ) // reverse order!
    {
	Vertex v = Vertex::linkedTriad( *verts[i], -1, m_point );
	v.normalize();
	array.add( m_vertTree.getVertex(v) );
    }


    newTriangles = Triangle::triangulate( array );
    newTriangles = cullDegenerate( newTriangles );
    assert( array.length() == 0 );

    // check that the result makes sense
    /*
    for( int i=0; i<newTriangles.length(); i++ )
    {
	Vertex n = newTriangles[i]->normal();
	assert( Vertex::dotProduct(n,m_normal) > 0 );
    }
    */
    
    //// clip triangles against existing occluders

    /*
    cout << "Sphere::render: " << m_triangles.length() 
	<< " triangles stored\n";
    */

#ifdef USE_TRITREE
    DArray<Triangle> clippedTriangles;
    for( int i=0; i<newTriangles.length(); i++ )
    {
	Triangle *t = newTriangles[i];
	DArray<Triangle> targets;
	DArray<Triangle> existing = m_tritree->getTriangles( *t );
	targets.add( t );
	
	for( int j=0; j<existing.length(); j++ )
	{
	    Triangle &t1 = *existing[j];
	    DArray<Triangle> newTargets;
	    for( int k=0; k<targets.length(); k++ )
	    {
		Triangle &t2 = *targets[k];
		ClipMatrix cp( t1, t2 );
		if( cp.overlap() )
		{
		    newTargets.append( subtract(m_vertTree,t2,t1) );
		    delete &t1;
		}
		else
		    newTargets.add( &t1 );
	    }
	    targets = newTargets;
	}
	clippedTriangles.append( targets );
    }
    newTriangles = clippedTriangles;

#else
    
    // newTriangles is now the triangulated input polygon
    // cut each existing triangle out from each candidate
    for( int i=0; i<m_triangles.length(); i++ )
    {
	Triangle &t1 = *m_triangles[i];
	DArray<Triangle> clippedTriangles;

	for( int j=0; j<newTriangles.length(); j++ )
	{
	    Triangle &t2 = *newTriangles[j];
	    ClipMatrix cp( t1, t2 );

	    if( cp.overlap() )
	    {
		clippedTriangles.append( subtract(m_vertTree,t2,t1) );
		delete newTriangles[j];
	    }
	    else
		clippedTriangles.add( newTriangles[j] );
	}
	newTriangles = clippedTriangles;  // newTriangles all accounted for

	// if all triangles have been consumed, we can stop
	if( newTriangles.length() == 0 )
	    break;
	/*
	cout << "  render: iteration: " << i << ": " << newTriangles.length()
	    << " triangles produced\n";
	    */
	// assertConsistent( newTriangles );
    }
#endif // USE_TRITREE
    
    /*
    cout << "\nSphere::render: " << newTriangles.length() 
	<< " triangles produced in all\n";
	*/
    return cullDegenerate( newTriangles );
}


void
Sphere::insert( DArray<Triangle> &a )
{
#ifdef USE_TRITREE
    m_tritree->add( a );
#else
    m_triangles.append( a );
#endif // USE_TRITREE
}
    
/////////////////////////////////////////////////////////////////////////////

// the arcs of 'a' are consumed!
DArray<Triangle>
subtract( VertexAVL &tree, Triangle &a, Triangle &b )
{
    DArray<Triangle> output;
    DArray<Arc> poly = a.arcs();

    // assertPolygon( a.arcs() );
    // assertPolygon( b.arcs() );
    a.dismantle();
    for( int i=0; i<3; i++ )
    {
	DArray<Arc> cropped = crop( tree, b.getArc(i), poly );
	if( cropped.length() )
	{
	    // assertPolygon( cropped );
	    DArray<Triangle> x = Triangle::triangulate( cropped );
	    // assertConsistent( x );
	    output.append( x );
	}

	if( !poly.length() )
	    break;
	else
	{
	    // assertPolygon( poly );
	}
    }
    poly.clear( 1 );
    return output;
}


// 'tree' is to make new vertices, 'arc' represents a great circle to clip
// against, 'poly' should be a convex polygon.
//
// returns the elements cropped away, modifies 'poly' to contain
// those elements 'inside' the cropping line
DArray<Arc>
crop( VertexAVL &tree, Arc &arc, DArray<Arc> &poly )
{
    assert( poly.length() > 2 );

    DArray<Arc> output;
    int *code = new int[poly.length()+1];
    int in = 0;
    int out = 0;

    // calcluate in/out codes, and count them
    // cout << "\ncrop:\n  " << arc << endl << "poly:\n" << poly;
    // cout << "  codes: ";
    for( int i=0; i<poly.length(); i++ )
    {
	code[i] = arc.inOrOut( poly[i]->getV1() );
	// cout << code[i] << " ";
	if( code[i] > 0 )
	    in++;
	else if( code[i] < 0 )
	    out++;
    }
    code[poly.length()] = code[0];
    // cout << '\n';

    // check degenerate cases
    if( out == 0 ) // it's all inside
    {
	// cout << "all inside\n";
	delete code;
	return output;
    }
    else if( in == 0 )
    {
	// cout << "all outside\n";
	delete code;
	output = poly;
	poly.clear();
	return output;
    }

    // hard case:  polygon is bisected by arc
    // cout << "intersection\n";
    DArray<Arc> newPoly; // this will store the inside polygon
    int gotFirst = 0;
    // assertPolygon( poly );
    ArcVertex *firstVertex;
    for( int i=0; i<poly.length(); i++ )
    {
	/*
	cout << i << ": begin: " << *poly[i];
	if( poly.length() > 3)
	    cout << 3 << ": " << *poly[3];
	    */
	assert( code[i] || code[i+1] );  // degenerate polygon passed?
	int sum = code[i] + code[i+1];

	if( sum > 1 ) // arc is inside
	{
	    // cout << i << ": arc inside\n";
	    newPoly.add( poly[i] );
	}
	else if( sum < -1 ) // arc is outside
	{
	    // cout << i << ": arc outside\n";
	    output.add( poly[i] );
	}
	else if( sum == 0 ) // one is outside, one is inside
	{
	    assert( gotFirst<2 );

	    // cout << i << ": outside/inside\n";
	    int inToOut = code[i] > 0;
	    assert( fabs(poly[i]->getV1().mag2()-1) < 0.05 );
	    assert( fabs(poly[i]->getV2().mag2()-1) < 0.05 );
	    Vertex _v = Arc::intersection( arc.getGreatCircle(), *poly[i] );
	    // cout << "crop: intersection: " << _v << '\n';
	    // assertIntersection( _v, arc.getGreatCircle(), *poly[i] );
	    ArcVertex &v = *tree.getVertex( _v );
	    Arc *firstHalf = new Arc( poly[i]->getV1(), v );
	    Arc *secondHalf = new Arc( v, poly[i]->getV2() );

	    delete poly[i]; // this arc must get destroyed

	    // add the first half to appropriate list
	    if( inToOut )
		newPoly.add( firstHalf );
	    else
		output.add( firstHalf );

	    // add the second half
	    if( gotFirst )
	    {
		if( inToOut )
		{
		    newPoly.add( new Arc(&v,firstVertex) );
		    output.add( new Arc(firstVertex,&v) );
		    output.add( secondHalf );
		}
		else
		{
		    output.add( new Arc(&v,firstVertex) );
		    newPoly.add( new Arc(firstVertex,&v) );
		    newPoly.add( secondHalf );
		}
	    }
	    else
	    {
		firstVertex = &v;
		if( inToOut )
		    output.add( secondHalf );
		else
		    newPoly.add( secondHalf );
	    }
	    gotFirst++;
	}
	else  // one is on arc, one is in/out
	{
	    assert( abs(sum) == 1 );
	    // cout << i << ": on/inout\n";

	    if( code[i+1] == 0 )  // the second is on the arc
	    {
		assert( gotFirst<2 );

		DArray<Arc> &array1 = code[i]==1 ? newPoly : output;
		DArray<Arc> &array2 = code[i]==1 ? output : newPoly;
		if( gotFirst )
		{
		    ArcVertex &v = poly[i]->getV2();
		    array1.add( poly[i] );

		    // array1.add( new Arc(firstVertex,&v) ); // wrong?
		    array1.add( new Arc(&v,firstVertex) );

		    // array2.add( new Arc(&v,firstVertex) ); // wrong?
		    array2.add( new Arc(firstVertex,&v) );
		}
		else
		{
		    array1.add( poly[i] );
		    firstVertex = &( poly[i]->getV2() );
		}
		gotFirst++;
	    }
	    else // the first is on the arc
	    {
		DArray<Arc> &array = code[i+1]==1 ? newPoly : output;
		array.add( poly[i] );
	    }
	}
    }
    poly = newPoly;

    assert( poly.length() > 2 );
    assert( output.length() > 2 );

    output = normalize( output );

    /*
    if( output.length() >= 3 )
	assertPolygon( output );
    */
    
    poly = normalize( poly );
    /*
    if( poly.length() >= 3 )
	assertPolygon( poly );
    */
    
    delete code;
    return output;
}


DArray<Triangle>
cullDegenerate( DArray<Triangle> &array )
{
    const double TOLERANCE = 1e-8;
    DArray<Triangle> newArray( array.length() );

    for( int i=0; i<array.length(); i++ )
    {
	Triangle *t = array[i];
	for( int j=0; j<3; j++ )
	{
	    int k = (j+1) % 3;
	    Vertex &v1 = t->getVertex( j );
	    Vertex &v2 = t->getVertex( k );
	    Vertex diff = v1 - v2;
	    if( diff.mag2() < TOLERANCE )
	    {
		/*
		cout << "Warning:  culling degenerate triangle\n";
		cout << "   v1: " << v1 << endl;
		cout << "   v2: " << v2 << endl;
		cout << "   mag2: " << diff.mag2() << endl;
		*/
		goto nextIter;
	    }
	}
	array[i] = 0;
	newArray.add( t );
    nextIter:
	t = 0;  // why does a label have to be followed by a statement?
    }
    array.clear( 1 );  // erase the degenerate ones
    return newArray;
}


// collapse arcs with same vertices
DArray<Arc>
normalize( DArray<Arc> &input )
{
    DArray<Arc> output;
    int printed = 0;

    for( int i=0; i<input.length(); i++ )
    {
	Arc *cur = input[i];
	if( &(cur->getV1()) == &(cur->getV2()) )
	{
	    /*
	    if( !printed )
	    {
		cout << "sphere.C:normalize():\n";
		for( int j=0; j<input.length(); j++ )
		    cout << '\t' << input[i]->getV1() << endl;
		printed = 1;
	    }
	    cout << "sphere.C:normalize():  Warning:  collapsing arc!\n";
	    */
	    delete input[i];
	}
	else
	    output.add( cur );
    }

    // check that there are still at least three
    if( output.length() < 3 )
    {
	// cout << "sphere.C:normalize():  Warning:  collapsing triangle!\n";
	output.clear( 1 );
    }
    return output;
}


// get in/out codes for t1 clipping t2 and t2 clipping t1
ClipMatrix::ClipMatrix( Triangle &_t1, Triangle &_t2 )
{
    Triangle *triangles[2] = { &_t1, &_t2 };

    for( int i=0; i<2; i++ )
    {
	Triangle &t1 = *triangles[i];
	Triangle &t2 = *triangles[(i+1)%2];
	for( int j=0; j<3; j++ )
	{
	    Arc &a = t1.getArc( j );
	    for( int k=0; k<3; k++ )
		m[i][j][k] = a.inOrOut( t2.getVertex(k) );
	}
    }
}


// if all vertices of one triangle are "outside" an edge of the other,
// then there is no overlap
int
ClipMatrix::overlap()
{
    for( int i=0; i<2; i++ )
    {
	for( int j=0; j<3; j++ )
	{
	    int *v = &m[i][j][0];
	    if( v[0] <= 0 && v[1] <= 0 && v[2] <= 0 )
		return 0;
	}
    }
    return 1;
}


void
assertPolygon( DArray<Arc> &arcs )
{
    int len = arcs.length();

    assert( len >= 3 );

    // check that arcs match up
    for( int i=0; i<len; i++ )
    {
	int j = (i+1) % len;
	ArcVertex *v1 = &(arcs[i]->getV2());
	ArcVertex *v2 = &(arcs[j]->getV1());
	assert( v1 == v2 );
    }

    // check that vertices are more or less normalized
    for( int i=0; i<len; i++ )
    {
	double delta = fabs( arcs[i]->getV1().mag2() - 1 );
	assert( delta < .05 );
    }
    
    // check that adjacent vertices are not the same
    for( int i=0; i<len; i++ )
    {
	ArcVertex *v1 = &(arcs[i]->getV1());
	ArcVertex *v2 = &(arcs[i]->getV2());
	assert( v1 != v2 );
    }

    // check that it's convex
    for( int i=0; i<len; i++ )
    {
	int j = (i+1) % len;
	Vertex &v1 = arcs[i]->getV1();
	Vertex &v2 = arcs[i]->getV2();
	Vertex &v3 = arcs[j]->getV2();

	Vertex d1 = Vertex::linkedTriad( v1, -1, v2 );
	Vertex d2 = Vertex::linkedTriad( v3, -1, v2 );

	d1.normalize();
	d2.normalize();

	Vertex c = Vertex::crossProduct( d2, d1 );
	if( asin(c.mag()) < 0 )
	{
	    cout << "Non-convex result (" << i << " -> " << i+1 << "):\n";
	    cout << arcs;
	    cout.flush();
	    assert( 0 );
	}
    }
}


ostream &
operator<<( ostream &out, DArray<Arc> &a )
{
    out << "DArray<Arc>:\n";
    for( int i=0; i<a.length(); i++ )
    {
	Arc &arc = *a[i];
	out << arc;
    }
    return out;
}


ostream &
operator<<( ostream &out, Arc &a )
{
    out << "  arc:" << a.getV1() << " -> " << a.getV2() << '\n';
    return out;
}


void
assertIntersection( Vertex &v, Vertex &gn, Arc &a )
{
#ifdef DEBUG    
    double e = 0.0001;
    Vertex &agn = a.getGreatCircle();
    Vertex &v1 = a.getV1();
    Vertex &v2 = a.getV2();
    double d = Vertex::dotProduct( v1, v2 );

    if( fabs(v.mag()-1) > e )
    {
	cout << "assertIntersection: v.mag: " << v.mag() << '\n';
	assert( 0 );
    }

    if( fabs(gn.mag()-1) > e )
    {
	cout << "assertIntersection: gn.mag: " << gn.mag() << '\n';
	assert( 0 );
    }

    if( fabs(Vertex::dotProduct(v,gn)) > e )
    {
	cout << "assertIntersection: v . gn: " << 
	    Vertex::dotProduct(v,gn) << '\n';
	assert( 0 );
    }

    if( Vertex::dotProduct(v,v1) < d || Vertex::dotProduct(v,v2) < d )
    {
	cout << "assertIntersection: v not between (v1,v2)\n" 
	    << "  v: " << v << "  v1: " << v1 << "  v2:" << v2 << '\n';
	assert( 0 );
    }
#endif // DEBUG    
}

void
assertConsistent( DArray<Triangle> &x )
{
#ifdef DEBUG
    // check that they have different vertices, and a good normal
    // Hee, hee, Dave Lesage:  "The good normal."
    for( int i=0; i<x.length(); i++ )
    {
	/*
	const double TOLERANCE = 1e-7;
	DArray<Arc> &arcs = x[i]->arcs();
	Vertex v1 = arcs[0]->getV1();
	Vertex v2 = arcs[1]->getV1();
	Vertex v3 = arcs[2]->getV1();

	Vertex diff = v1 - v2;
	assert( diff.mag2() > TOLERANCE );
	diff = v2 - v3;
	assert( diff.mag2() > TOLERANCE );
	diff = v3 - v1;
	assert( diff.mag2() > TOLERANCE );
	*/
	
	Vertex n = x[i]->normal();
	assert( n[0] < 1.05  && n[1] < 1.05 && n[2] < 1.05 );
    }

    // cout << "assertConsistent: checking " << x.length() << " triangles\n";
    // assert that they probably face the same way
    for( int i=0; i<x.length(); i++ ) // for every triangle
    {
	DArray<Arc> &arcs1 = x[i]->arcs();
	for( int j=0; j<arcs1.length(); j++ ) // for each arc of the triangle
	{
	    Arc &a1 = *arcs1[j];

	    for( int k=0; k<x.length(); k++ ) // for ever other triangle
	    {
		if( i == k ) // don't test same triangles
		    continue;
		DArray<Arc> &arcs2 = x[k]->arcs();
		for( int l=0; l<arcs2.length(); l++ ) // for every arc of that
		{
		    Arc &a2 = *arcs2[l];

		    myAssert( &(a1.getV1()) != &(a2.getV1()) ||
			&(a1.getV2()) != &(a2.getV2()) );
		}
	    }
	}
    }

    // check that they have normalized vertices
    for( int i=0; i<x.length(); i++ )
    {
	Triangle &t = *x[i];
	for( int j=0; j<3; j++ )
	{
	    double delta = fabs( t.getVertex(j).mag2() - 1 );
	    assert( delta < 0.5 );
	}
    }
#endif // DEBUG
}


// this is a hack necessitated by a bug in gdb that makes it impossible
// to view local variables in a function with a failed assertion.
void
myAssert( int cond )
{
    assert( cond );
}
