#include <math.h>
#include <assert.h>
#include "tritree.h"


// these outcodes are a little weird.  To figure out the codes,
// draw a diagram as indicated by the algorithm in the constructor, and
// the definition of the arcs delimiting the partitions.  The second sector
// is the one "inside" arc 2 and "outside" arc 1.
class OutCodes
{
    unsigned int m_code;

    static const unsigned int m_triCont[];
    static const unsigned int m_triContMask[];
    static const unsigned int m_quadCont[];
    static const unsigned int m_quadContMask[];

    int match( unsigned int c, unsigned int m )
    { return ( m_code & m ) == c; }

public:
    OutCodes( DArray<Arc> &, Triangle & );

    //// queries

    // true if triangle is entirely contained in the partition
    int triContained( int i )
    {
	return match( m_triCont[i], m_triContMask[i] );
    }

    // similar
    int quadContained( int i )
    {
	return match( m_quadCont[i], m_quadContMask[i] );
    }

    // true if triangle overlaps the partition - conservative
    // ex: it doesn't overlap partition one if either it's all outside
    //     arc 1, or all inside arc 3.
    int triIntersected( int i )
    {
	int j = (--i + 2) % 3;
	return !( match(0,m_triCont[i]) || match(m_triCont[j],m_triCont[j]) );
    }

    // similar
    int quadIntersected( int );
};

const unsigned int OutCodes::m_triCont[] = { 7, 56, 448 };
const unsigned int OutCodes::m_triContMask[] = { 455, 63, 504 };
const unsigned int OutCodes::m_quadCont[] = { 63, 56, 0, 7 };
const unsigned int OutCodes::m_quadContMask[] = { 63, 63, 63, 63 };

/////////////////////////////////////////////////////////////////////////////
////////////////////////////  TriTreeNode  //////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


TriTreeNode::TriTreeNode( ArcVertex &v1, ArcVertex &v2, ArcVertex &v3,
			  VertexAVL &tree, int count )
{
    if( count )
    {
	Vertex _c = v1 + v2 + v3;
	_c.normalize();
	ArcVertex *c = tree.getVertex( _c );
	
	m_arcs.add( new Arc(&v1,c) );
	m_arcs.add( new Arc(&v2,c) );
	m_arcs.add( new Arc(&v3,c) );
	count--;

	m_children.add( new TriTreeNode(v1,v2,*c,tree,count) );
	m_children.add( new TriTreeNode(v2,v3,*c,tree,count) );
	m_children.add( new TriTreeNode(v3,v1,*c,tree,count) );		
    }
}

TriTreeNode::~TriTreeNode()
{
    m_triangles.clear( 1 ); // delete triangles
    m_arcs.clear( 1 );  // delete these, too
    m_children.clear( 1 ); // and delete these, too
}


void
TriTreeNode::add( Triangle &t )
{

    if( m_arcs.length() )
    {
	OutCodes oc( m_arcs, t );
	if( oc.triContained(0)  )
	    m_children[0]->add( t );
	else if( oc.triContained(1) )
	    m_children[1]->add( t );
	else if( oc.triContained(2) )
	    m_children[2]->add( t );
	else
	    m_triangles.add( &t );
    }
    else
	m_triangles.add( &t );
}


void
TriTreeNode::getTriangles( DArray<Triangle> &a, Triangle &t )
{
    a.append( m_triangles );

    if( m_arcs.length() )
    {
	OutCodes oc( m_arcs, t );
	if( oc.triIntersected(1) )
	    m_children[0]->getTriangles( a, t );
	if( oc.triIntersected(2) )
	    m_children[1]->getTriangles( a, t );
	if( oc.triIntersected(3) )
	    m_children[2]->getTriangles( a, t );
    }
}


void
TriTreeNode::getAllTriangles( DArray<Triangle> &a )
{
    a.append( m_triangles );
    for( int i=0; i<m_children.length(); i++ )
	m_children[i]->getAllTriangles( a );
}


/////////////////////////////////////////////////////////////////////////////
//////////////////////////////  TriTree  ////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

TriTree::TriTree( Vertex n, int count )
{
    assert( count > 0 );
    
    // find a couple of vectors orthogonal to normal
    Vertex v1, v2, v3, v4;
    int i;
    
    if( fabs(n[0]) > EPSILON )
	i = 0;
    else if( fabs(n[1]) > EPSILON )
	i = 1;
    else
	i = 2;
    int j = (i+1) % 3;
    int k = (i+2) % 3;
    v1[i] = -(n[j] + n[k])/n[i];
    v1[j] = 1;
    v1[k] = 1;
    v1.normalize();
    v2 = n ^ v1;
    v3 = -v1;
    v4 = -v2;

    ArcVertex *c1 = m_verts.getVertex( v1 );
    ArcVertex *c2 = m_verts.getVertex( v2 );
    ArcVertex *c3 = m_verts.getVertex( v3 );
    ArcVertex *c4 = m_verts.getVertex( v4 );
    m_arcs.add(	new Arc(c1,c3) );
    m_arcs.add(	new Arc(c2,c4) );

    count--;
    ArcVertex *c = m_verts.getVertex( n );
    m_children.add( new TriTreeNode(*c1,*c2,*c,m_verts,count) );
    m_children.add( new TriTreeNode(*c2,*c3,*c,m_verts,count) );
    m_children.add( new TriTreeNode(*c3,*c4,*c,m_verts,count) );
    m_children.add( new TriTreeNode(*c4,*c1,*c,m_verts,count) );
}



TriTree::~TriTree()
{
    m_arcs.clear( 1 );  // delete
    m_children.clear( 1 ); // delete
    m_triangles.clear( 1 );

    DSArray<Vertex> items = m_verts.items();
    if( items.length() != 0 )
    {
	cout << items.length() << " More ArcVertices undeleted\n";
	for( int i=0; i<items.length(); i++ )
	{
	    ArcVertex *v = (ArcVertex *) items[i];
	    assert( v->count() == 0 );
	    delete v;
	}
    }
}


void
TriTree::add( Triangle &t )
{
    OutCodes oc( m_arcs, t );

    for( int i=0; i<3; i++ )
    	if( oc.quadContained(i) )
	{
	    m_children[i]->add( t );
	    return;
	}

    m_triangles.add( &t );
}


void
TriTree::add( DArray<Triangle> &a )
{
    for( int i=0; i<a.length(); i++ )
	add( *a[i] );
}


DArray<Triangle>
TriTree::getTriangles( Triangle &t )
{
    DArray<Triangle> a = m_triangles;
    OutCodes oc( m_arcs, t );

    for( int i=0; i<3; i++ )
	if( oc.quadIntersected(i) )
	    m_children[i]->getTriangles( a, t);
}


DArray<Triangle>
TriTree::getAllTriangles()
{
    DArray<Triangle> a = m_triangles;

    for( int i=0; i<4; i++ )
	m_children[i]->getAllTriangles( a );
}


/////////////////////////////////////////////////////////////////////////////
//////////////////////////////  OutCodes  ///////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


OutCodes::OutCodes( DArray<Arc> &arcs, Triangle &t )
{
    m_code = 0;
    
    for( int i=0; i<arcs.length(); i++ )
    {
	for( int j=0; j<3; j++ )
	{
	    int code = arcs[i]->inOrOut( t.getVertex(j) );

	    if( code > 0 ) // conservative
		code |= 1 << ( i*3 + j );
	}
    }
}
		

int
OutCodes::quadIntersected( int i )
{
    assert( i >= 0 && i < 4 );
    
    switch( i )
    {
    case 0:
	return !( match(0,m_triCont[0]) ||
		  match(0,m_triCont[1]) );
    case 1:
	return !( match(m_triCont[0],m_triCont[0]) ||
		  match(0,m_triCont[1]) );
    case 2:
	return !( match(m_triCont[1],m_triCont[1]) ||
		  match(m_triCont[0],m_triCont[0]) );
    case 3:
	return !( match(m_triCont[1],m_triCont[1]) ||
		  match(0,m_triCont[0]) );
    }
    return 0; // shouldn't be reached
}
